import numpy as np

from act.act_types import SimulationConstants


class PospischilsPY:
    # Cell
    cell_hoc_file = "../data/Pospischil/sPY/template.hoc"
    cell_name = "sPY"

    # Simulation parameters
    h_v_init = -67.0  # (mV)
    h_tstop = 2000  # (ms)
    h_i_delay = 500  # (ms)
    h_i_dur = 1500  # (ms)
    h_dt = 0.025

    # Optimization parameters
    amps = list(np.arange(-2, 10, 0.1))
    params = ["g_pas", "gnabar_hh2", "gkbar_hh2", "gkbar_im"]
    lows = [5.0e-05, 2.5e-02, 2.5e-03, 3.5e-05]
    highs = [0.0002, 0.1, 0.01, 0.00014]

    # Summary features
    spike_threshold = 20  # (mV)

    # Target-sim match conditions (max abs diff between sim and target)
    mc_num_spikes = 1
    mc_interspike_time = 200  # (ms)
    mc_min_v = 1  # (mV)
    mc_mean_v = 2  # (mV)
    mc_max_v = 1  # (mV)

    # Segregation
    segr_param_inds = [[0], [1, 2, 3]]
    segr_voltage_bounds = [[-100, -65], [-65, 100]]
    segr_time_bounds = [[0, 500], [0, 2000]]

    # Target voltage
    target_V = None
    target_params = [0.0001, 0.05, 0.005, 7e-5]

    # Runtime
    run_mode = "segregated"  # "original", "segregated"
    modfiles_mode = "segregated"  # Used only for the output folder name
    modfiles_folder = "../data/Pospischil/sPY/seg_modfiles"
    num_repeats = 3
    num_amps_to_match = 12
    num_epochs = 5000

    # Output
    output_folder = "output_Pospischil_sPY"
    produce_plots = True


class PospischilsPYr:
    # Cell
    cell_hoc_file = "../data/Pospischil/sPYr/template.hoc"
    cell_name = "sPYr"

    # Simulation parameters
    h_v_init = -67.0  # (mV)
    h_tstop = 2000  # (ms)
    h_i_delay = 500  # (ms)
    h_i_dur = 1500  # (ms)
    h_dt = 0.025

    # Optimization parameters
    amps = list(np.arange(-2, 10, 0.1))
    params = ["g_pas", "gnabar_hh2", "gkbar_hh2", "gkbar_im", "gcabar_it"]
    lows = [5.0e-06, 2.5e-02, 2.5e-03, 1.5e-05, 5.0e-04]
    highs = [2.0e-05, 1.0e-01, 1.0e-02, 6.0e-05, 2.0e-03]

    # Summary features
    spike_threshold = 20  # (mV)

    # Target-sim match conditions (max abs diff between sim and target)
    mc_num_spikes = 1
    mc_interspike_time = 200  # (ms)
    mc_min_v = 1  # (mV)
    mc_mean_v = 2  # (mV)
    mc_max_v = 1  # (mV)

    # Segregation
    segr_param_inds = [[0], [1, 2, 3, 4]]
    segr_voltage_bounds = [[-100, -65], [-65, 100]]
    segr_time_bounds = [[0, 500], [0, 2000]]

    # Target voltage
    target_V = None
    target_params = [1e-5, 0.05, 0.005, 3e-5, 0.001]

    # Runtime
    run_mode = "segregated"  # "original", "segregated"
    modfiles_mode = "segregated"  # Used only for the output folder name
    modfiles_folder = "../data/Pospischil/sPYr/seg_modfiles"
    num_repeats = 3
    num_amps_to_match = 12
    num_epochs = 5000

    # Output
    output_folder = "output_Pospischil_sPYr"
    produce_plots = True


pospischilsPYr: SimulationConstants = {
    "cell": {
        "cell_hoc_file": "../data/Pospischil/sPYr/template.hoc",
        "cell_name": "sPYr",
    },
    "simulation_parameters": {
        "h_v_init": -67.0,  # (mV)
        "h_tstop": 2000,  # (ms)
        "h_i_delay": 500,  # (ms)
        "h_i_dur": 1500,  # (ms)
        "h_dt": 0.025,
    },
    "optimization_parameters": {
        "amps": list(np.arange(-2, 10, 0.1)),
        "params": [
            {"channel": "g_pas", "low": 5.0e-06, "high": 2.0e-05},
            {"channel": "gnabar_hh2", "low": 2.5e-02, "high": 1.0e-01},
            {"channel": "gkbar_hh2", "low": 2.5e-03, "high": 1.0e-02},
            {"channel": "gkbar_im", "low": 1.5e-05, "high": 6.0e-05},
            {"channel": "gcabar_it", "low": 5.0e-04, "high": 2.0e-03},
        ],
    },
    "summary_features": {
        "spike_threshold": 20,  # (mV)
        # Target-sim match conditions (max abs diff between sim and target)
        "mc_num_spikes": 1,
        "mc_interspike_time": 200,  # (ms)
        "mc_min_v": 1,  # (mV)
        "mc_mean_v": 2,  # (mV)
        "mc_max_v": 1,  # (mV)
    },
    "segregation": {
        "segr_param_inds": [[0], [1, 2, 3, 4]],
        "segr_voltage_bounds": [[-100, -65], [-65, 100]],
        "segr_time_bounds": [[0, 500], [0, 2000]],
        # Target voltage
        "target_V": None,
        "target_params": [1e-5, 0.05, 0.005, 3e-5, 0.001],
    },
    "output": {"output_folder": "output_Pospischil_sPYr", "produce_plots": True},
    "run_mode": "segregated",  # "original", "segregated"
    "modfiles_mode": "segregated",  # Used only for the output folder name
    "modfiles_folder": "../data/Pospischil/sPYr/seg_modfiles",
    "num_repeats": 3,
    "num_amps_to_match": 12,
    "num_epochs": 5000,
}



import os

import matplotlib.pyplot as plt
import numpy as np
import torch

from act.act_types import SimulationConstants
from act.metrics import correlation_score, mse_score
from act.optim import ACTOptimizer


def save_prediction_plots(
    target_V: torch.Tensor,
    amp: list,
    simulation_constants: SimulationConstants,
    predicted_params_values: torch.Tensor,
    output_folder: str,
) -> None:
    _, ax = plt.subplots(1, 1, figsize=(10, 10))
    optim = ACTOptimizer(simulation_constants=simulation_constants)
    params = [
        p["channel"] for p in simulation_constants["optimization_parameters"]["params"]
    ]
    simulated_data = optim.simulate(
        amp, params, predicted_params_values.detach().numpy()
    )
    simulated_data = optim.resample_voltage(
        V=simulated_data.reshape((1, -1)), num_obs=target_V.shape[1]
    )

    title = f"I = {amp} nA"
    ax.plot(simulated_data.flatten(), label="Simulated")
    ax.plot(target_V.flatten(), label="Target")
    ax.set_title(title)
    ax.set_xlabel("Timestamp")
    ax.set_ylabel("V (mV)")
    ax.legend()
    ax.grid()

    plt.savefig(os.path.join(output_folder, f"{amp * 1000}nA.png"))
    plt.close()


def save_mse_corr(
    target_V: torch.Tensor,
    simulation_constants: SimulationConstants,
    predicted_params_values: torch.Tensor,
    output_folder: str,
) -> None:
    with open(os.path.join(output_folder, "metrics.csv"), "w") as file:
        file.write(f"amp,mse,corr\n")

    optim = ACTOptimizer(simulation_constants=simulation_constants)
    for ind, amp in enumerate(simulation_constants["optimization_parameters"]["amps"]):
        params = [
            p["channel"]
            for p in simulation_constants["optimization_parameters"]["params"]
        ]
        sim_data = optim.simulate(amp, params, predicted_params_values.detach().numpy())
        simulated_data = optim.resample_voltage(
            V=sim_data.reshape((1, -1)), num_obs=target_V.shape[1]
        )
        mse = mse_score(target_V[ind].reshape(-1, 1), simulated_data.reshape(-1, 1))
        corr = correlation_score(
            target_V[ind].reshape(1, -1), simulated_data.reshape(1, -1)
        )

        with open(os.path.join(output_folder, "metrics.csv"), "a") as file:
            file.write(f"{amp},{mse},{corr}\n")



"""
These types should be used to define the user supplied simulation config/constants.

"""

from typing import List, TypedDict


class Cell(TypedDict):
    cell_hoc_file: str
    cell_name: str


class SimulationParameters(TypedDict):
    h_v_init: float
    h_tstop: int  # (ms)
    h_i_delay: int  # (ms)
    h_i_dur: int  # (ms)
    h_dt: float


class OptimizationParam(TypedDict):
    param: str
    low: float
    high: float


class OptimizationParameters(TypedDict):
    amps: List[float]
    params: List[OptimizationParam]


class SummaryFeatures:
    spike_threshold: int  # (mV)

    # Target-sim match conditions (max abs diff between sim and target)
    mc_num_spikes: int
    mc_interspike_time: int  # (ms)
    mc_min_v: int  # (mV)
    mc_mean_v: int  # (mV)
    mc_max_v: int  # (mV)


class Segregation(TypedDict):
    segr_param_inds: List[List[int]]
    segr_voltage_bounds: List[List[int]]
    segr_time_bounds: List[List[int]]

    # Target voltage
    target_V: int
    target_params: List[float]


class Output(TypedDict):
    output_folder: str
    produce_plots: bool


class SimulationConstants(TypedDict):
    cell: Cell
    simulation_parameters: SimulationParameters
    optimization_parameters: OptimizationParameters
    summary_features: SummaryFeatures
    segregation: Segregation
    output: Output

    run_mode: str  # "original", "segregated"
    modfiles_mode: str  # Used only for the output folder name
    modfiles_folder: str
    num_repeats: int
    num_amps_to_match: int
    num_epochs: int



import numpy as np
import torch
from neuron import h
from scipy.signal import resample

from act.act_types import SimulationConstants
from act.cell_model import CellModel
from act.logger import ACTDummyLogger


class ACTOptimizer:
    def __init__(
        self,
        simulation_constants: SimulationConstants,
        logger: object = None,
        reset_cell_params_to_lower_bounds_on_init: bool = True,
    ):
        self.constants = simulation_constants

        # Initialize standard run
        h.load_file("stdrun.hoc")

        # Initialize the cell
        self.cell = CellModel(
            hoc_file=self.constants["cell"]["cell_hoc_file"],
            cell_name=self.constants["cell"]["cell_name"],
        )
        if reset_cell_params_to_lower_bounds_on_init:
            params = [
                p["channel"]
                for p in self.constants["optimization_parameters"]["params"]
            ]
            lows = [
                p["low"] for p in self.constants["optimization_parameters"]["params"]
            ]
            self.cell.set_parameters(params, lows)

        # For convenience
        self.update_param_vars()

        # Set the logger
        self.logger = logger
        if self.logger is None:
            self.logger = ACTDummyLogger()

    def optimize(self, target_V: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def optimize_with_segregation(
        self, target_V: torch.Tensor, segregate_by: str
    ) -> torch.Tensor:
        raise NotImplementedError

    def cut_voltage_region(
        self, target_V: torch.Tensor, voltage_bounds: list
    ) -> torch.Tensor:
        cut_target_V = torch.zeros_like(target_V)

        for i in range(cut_target_V.shape[0]):
            trace = target_V[i]
            trace = trace[(trace >= voltage_bounds[0]) & (trace < voltage_bounds[1])]
            cut_target_V[i] = torch.tensor(
                resample(x=trace, num=target_V.shape[1])
            ).float()

        return cut_target_V

    def cut_time_region(
        self, target_V: torch.Tensor, time_bounds: tuple
    ) -> torch.Tensor:
        cut_target_V = torch.zeros_like(target_V)

        for i in range(cut_target_V.shape[0]):
            trace = target_V[i, time_bounds[0] : time_bounds[1]]
            cut_target_V[i] = torch.tensor(
                resample(x=trace, num=target_V.shape[1])
            ).float()

        return cut_target_V

    def simulate(
        self, amp: float, parameter_names: list, parameter_values: np.ndarray
    ) -> torch.Tensor:
        h.dt = self.constants["simulation_parameters"]["h_dt"]
        h.tstop = self.constants["simulation_parameters"]["h_tstop"]
        h.v_init = self.constants["simulation_parameters"]["h_v_init"]

        self.cell.set_parameters(parameter_names, parameter_values)
        self.cell.apply_current_injection(
            amp,
            self.constants["simulation_parameters"]["h_i_dur"],
            self.constants["simulation_parameters"]["h_i_delay"],
        )

        h.run()

        return torch.Tensor(self.cell.Vm.as_numpy()[:-1])

    def resample_voltage(self, V: torch.Tensor, num_obs: int) -> torch.Tensor:
        resampled_data = []
        for i in range(V.shape[0]):
            resampled_data.append(resample(x=V[i], num=num_obs))
        resampled_data = torch.tensor(np.array(resampled_data)).float()
        return resampled_data

    def update_param_vars(self) -> None:
        self.num_ampl = len(self.constants["optimization_parameters"]["amps"])
        self.num_params = len(self.constants["optimization_parameters"]["params"])


class GeneralACTOptimizer(ACTOptimizer):
    def __init__(
        self,
        simulation_constants: SimulationConstants,
        logger: object = None,
        reset_cell_params_to_lower_bounds_on_init: bool = True,
    ):
        super().__init__(
            simulation_constants=simulation_constants,
            logger=logger,
            reset_cell_params_to_lower_bounds_on_init=reset_cell_params_to_lower_bounds_on_init,
        )

        self.model = None
        self.model_pool = None

    def optimize(self, target_V: torch.Tensor) -> torch.Tensor:
        # Get voltage with characteristics similar to target
        simulated_V_for_next_stage, param_samples_for_next_stage = self.match_voltage(
            target_V
        )

        self.model = self.init_nn_model(
            in_channels=target_V.shape[1], out_channels=self.num_params
        )

        # Resample to match the length of target data
        resampled_data = self.resample_voltage(
            simulated_V_for_next_stage, target_V.shape[1]
        )

        # Train model
        self.train_model(resampled_data, param_samples_for_next_stage)

        # Predict and take max across ci to prevent underestimating
        predictions = self.predict_with_model(target_V)
        predictions = torch.max(predictions, dim=0).values

        return predictions

    def optimize_with_segregation(
        self, target_V: torch.Tensor, segregate_by: str = "voltage"
    ) -> torch.Tensor:
        if segregate_by == "voltage":
            cut_func = self.cut_voltage_region
            cut_bounds_list = self.constants["segregation"]["segr_voltage_bounds"]
        elif segregate_by == "time":
            cut_func = self.cut_time_region
            cut_bounds_list = self.constants["segregation"]["segr_time_bounds"]
        else:
            raise ValueError("segregate_by must be either 'voltage' or 'time'.")

        # Get voltage with characteristics similar to target
        simulated_V_for_next_stage, param_samples_for_next_stage = self.match_voltage(
            target_V
        )

        # Resample to match the length of target data
        resampled_data = self.resample_voltage(
            simulated_V_for_next_stage, target_V.shape[1]
        )

        # Store original params to restore later
        orig_params = [
            p["channel"] for p in self.constants["optimization_parameters"]["params"]
        ]
        orig_lows = [
            p["low"] for p in self.constants["optimization_parameters"]["params"]
        ]
        orig_highs = [
            p["high"] for p in self.constants["optimization_parameters"]["params"]
        ]

        self.model_pool = []
        prediction_pool = []
        for param_ind, bounds in zip(
            self.constants["segregation"]["segr_param_inds"], cut_bounds_list
        ):
            # Set the parameters
            params = [orig_params[v] for v in param_ind]
            lows = [orig_lows[v] for v in param_ind]
            highs = [orig_highs[v] for v in param_ind]
            self.num_params = len(params)

            # Cut the required segment
            cut_target_V = cut_func(resampled_data, bounds)

            # Train a model for this segment
            self.model = self.init_nn_model(
                in_channels=cut_target_V.shape[1], out_channels=len(param_ind)
            )
            self.train_model(
                cut_target_V,
                param_samples_for_next_stage[:, param_ind],
                params=params,
                lows=lows,
                highs=highs,
            )

            # Predict and take max across ci to prevent underestimating
            predictions = self.predict_with_model(cut_target_V, lows, highs)
            predictions = torch.max(predictions, dim=0).values

            # Update cell model
            self.cell.set_parameters(params, predictions)

            # Update the pools
            prediction_pool.append(predictions)
            self.model_pool.append(self.model)

        prediction_pool = torch.cat(prediction_pool, dim=0)

        return prediction_pool

    def init_nn_model(self, in_channels: int, out_channels: int) -> torch.nn.Sequential:
        model = torch.nn.Sequential(
            torch.nn.Linear(in_channels, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, out_channels),
            torch.nn.Sigmoid(),
        )
        return model

    def train_model(
        self,
        voltage_data: torch.Tensor,
        target_params: torch.Tensor,
        params,
        lows,
        highs,
    ) -> None:
        optim = torch.optim.SGD(self.model.parameters(), lr=1e-8)
        sigmoid_mins = torch.tensor(lows)
        sigmoid_maxs = torch.tensor(highs)

        self.logger.info("Training a model with SGD optimizer and lr = 1e-8.")
        self.logger.info(
            f"Number of trainable parameters: {sum(p.numel() for p in self.model.parameters() if p.requires_grad)}"
        )

        self.model.train(True)
        for i in range(voltage_data.shape[0]):
            # Track loss to stop if it starts to go up
            loss0 = np.inf

            for ne in range(self.constants["num_epochs"]):
                pred = (
                    self.model(voltage_data[i]) * (sigmoid_maxs - sigmoid_mins)
                    + sigmoid_mins
                )
                loss = torch.nn.functional.l1_loss(pred, target_params[i])

                if ne % 100 == 0:
                    self.logger.epoch(ne, "l1_loss", float(loss.detach().numpy()))

                if loss.detach().numpy() >= loss0:
                    break
                loss0 = loss.detach().numpy()

                loss.backward()
                optim.step()

    def predict_with_model(self, target_V: torch.Tensor, lows, highs) -> torch.Tensor:
        sigmoid_mins = torch.tensor(lows)
        sigmoid_maxs = torch.tensor(highs)

        self.model.eval()
        outs = []
        for i in range(target_V.shape[0]):
            out = self.model(target_V[i]) * (sigmoid_maxs - sigmoid_mins) + sigmoid_mins
            outs.append(out.reshape(1, -1))

        return torch.cat(outs, dim=0)

    def match_voltage(self, target_V: torch.Tensor) -> tuple:
        # Get target voltage summary features
        num_target_spikes, target_interspike_times = self.extract_summary_features(
            target_V
        )  # (N x 1)

        # Set a vector of non-correspondence to target features
        non_corresp = torch.ones(self.num_ampl)

        simulated_V_for_next_stage = torch.zeros(
            (
                self.num_ampl,
                int(
                    self.constants["simulation_parameters"]["h_tstop"]
                    / self.constants["simulation_parameters"]["h_dt"]
                ),
            )
        )
        param_samples_for_next_stage = torch.zeros((self.num_ampl, self.num_params))

        self.logger.info(f"Matching {self.constants['num_amps_to_match']} amplitudes.")

        params = [
            p["channel"] for p in self.constants["optimization_parameters"]["params"]
        ]
        lows = [p["low"] for p in self.constants["optimization_parameters"]["params"]]
        highs = [p["high"] for p in self.constants["optimization_parameters"]["params"]]

        while torch.sum(non_corresp) > (
            self.num_ampl - self.constants["num_amps_to_match"]
        ):
            # Simulate with these samples
            simulated_V = torch.zeros(
                (
                    self.num_ampl,
                    int(
                        self.constants["simulation_parameters"]["h_tstop"]
                        / self.constants["simulation_parameters"]["h_dt"]
                    ),
                )
            )
            for ind, amp in enumerate(
                self.constants["optimization_parameters"]["amps"]
            ):
                if non_corresp[ind] == 1:
                    # For each current injection amplitude, sample random parameters
                    param_samples = np.random.uniform(low=lows, high=highs)
                    simulated_V[ind] = self.simulate(amp, params, param_samples)
                    param_samples_for_next_stage[ind] = torch.Tensor(param_samples)

            # Compute summary features for simulated data
            (
                num_spikes_simulated,
                simulated_interspike_times,
            ) = self.extract_summary_features(simulated_V)

            # Save those voltage traces which match the target
            cond_spikes = self.get_match_condition(
                num_target_spikes,
                num_spikes_simulated,
                self.constants["summary_features"]["mc_num_spikes"],
            )
            cond_times = self.get_match_condition(
                target_interspike_times,
                simulated_interspike_times,
                self.constants["summary_features"]["mc_interspike_time"],
            )
            cond_min = self.get_match_condition(
                torch.min(target_V, dim=1).values,
                torch.min(simulated_V, dim=1).values,
                self.constants["summary_features"]["mc_min_v"],
            )
            cond_mean = self.get_match_condition(
                torch.mean(target_V, dim=1),
                torch.mean(simulated_V, dim=1),
                self.constants["summary_features"]["mc_mean_v"],
            )
            cond_max = self.get_match_condition(
                torch.max(target_V, dim=1).values,
                torch.max(simulated_V, dim=1).values,
                self.constants["summary_features"]["mc_max_v"],
            )

            cond = cond_spikes & cond_times & cond_mean & cond_min & cond_max
            simulated_V_for_next_stage[cond] = simulated_V[cond]
            non_corresp[cond] = 0
            self.logger.info(
                f"Total amplitudes matched: {int(torch.sum(1 - non_corresp))}/{self.constants['num_amps_to_match']}."
            )

        simulated_V_for_next_stage = simulated_V_for_next_stage[non_corresp == 0]
        param_samples_for_next_stage = param_samples_for_next_stage[non_corresp == 0]
        inds_of_ampl_next_stage = np.where(non_corresp == 0)[0].tolist()

        self.logger.info(
            f"Matched amplitudes: {np.array(self.constants['optimization_parameters']['amps'])[inds_of_ampl_next_stage]}"
        )

        return simulated_V_for_next_stage, param_samples_for_next_stage

    def get_match_condition(
        self, target: torch.Tensor, simulated: torch.Tensor, threshold: float
    ):
        return (torch.abs(target - simulated) < threshold).flatten()

    def extract_summary_features(self, V: torch.Tensor) -> tuple:
        threshold_crossings = torch.diff(
            V > self.constants["summary_features"]["spike_threshold"], dim=1
        )
        num_spikes = torch.round(torch.sum(threshold_crossings, dim=1) * 0.5)

        interspike_times = torch.zeros((V.shape[0], 1))
        for i in range(threshold_crossings.shape[0]):
            interspike_times[i, :] = torch.mean(
                torch.diff(
                    torch.arange(threshold_crossings.shape[1])[
                        threshold_crossings[i, :]
                    ]
                ).float()
            )

        return num_spikes, interspike_times



import numpy as np
import torch

from act.act_types import SimulationConstants
from act.optim import ACTOptimizer


def get_voltage_trace_from_params(
    simulation_constants: SimulationConstants,
) -> torch.Tensor:
    optim = ACTOptimizer(simulation_constants=simulation_constants)
    target_V = []
    for amp in simulation_constants["optimization_parameters"]["amps"]:
        params = [
            p["channel"]
            for p in simulation_constants["optimization_parameters"]["params"]
        ]
        target_params = simulation_constants["segregation"]["target_params"]
        target_V.append(optim.simulate(amp, params, target_params).reshape(1, -1))
    target_V = torch.cat(target_V, dim=0)
    return target_V



import os

import numpy as np
import pandas as pd

output_folders = [
    "output_Pospischil_sPYr/original_original",
    "output_Pospischil_sPYr/segregated_segregated",
]

target_params = np.array([1e-5, 0.05, 0.005, 3e-5, 0.001])


def main():
    for output_folder in output_folders:
        metrics = pd.read_csv(os.path.join(output_folder, "metrics.csv"))
        preds = np.array(
            pd.read_csv(os.path.join(output_folder, "pred.csv"), index_col=0)
        )
        print(output_folder, ":")
        print(f"Med MSE: {metrics['mse'].median()} ({metrics['mse'].std()})")
        print(f"Med Corr: {metrics['corr'].median()} ({metrics['corr'].std()})")
        print(f"Pred MAE: {np.mean(np.abs(target_params - preds))}")
        print("----------\n")


if __name__ == "__main__":
    main()



import datetime
import json
import os
import sys
from multiprocessing import Process

import numpy as np
import pandas as pd
import torch
from neuron import h

from act.act_types import SimulationConstants
from act.analysis import save_mse_corr, save_prediction_plots
from act.logger import ACTLogger
from act.metrics import correlation_score, mse_score
from act.optim import GeneralACTOptimizer
from act.target_utils import get_voltage_trace_from_params


def _run(constants: SimulationConstants):
    if constants["num_epochs"] < 1000:
        raise ValueError("Number of epochs is expected to be >= 1000.")

    output_folder = constants["output"]["output_folder"]
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    os.system(f"nrnivmodl {constants['modfiles_folder']}")

    logger = ACTLogger()
    logger.info(
        f"Number of amplitudes: {len(constants['optimization_parameters']['amps'])}"
    )

    try:
        h.nrn_load_dll("./x86_64/.libs/libnrnmech.so")
    except:
        logger.info("Mod files already loaded. Continuing.")

    # Get target voltage
    if constants["segregation"]["target_V"] is not None:
        target_V = constants["segregation"]["target_V"].copy()
    elif constants["segregation"]["target_params"] is not None:
        target_V = get_voltage_trace_from_params(constants)
    else:
        raise ValueError

    logger.info(f"Target voltage shape: {target_V.shape}")

    # Run the optimizer
    pred_pool = []
    err_pool = []
    params = [p["channel"] for p in constants["optimization_parameters"]["params"]]
    for _ in range(constants["num_repeats"]):
        if constants["run_mode"] == "original":
            optim = GeneralACTOptimizer(simulation_constants=constants, logger=logger)
            predictions = optim.optimize(target_V)
        elif constants["run_mode"] == "segregated":
            optim = GeneralACTOptimizer(simulation_constants=constants, logger=logger)
            predictions = optim.optimize_with_segregation(target_V, "voltage")
        else:
            raise ValueError

        pred_pool.append(predictions)

        sims = []
        for amp in constants["optimization_parameters"]["amps"]:
            sims.append(
                optim.simulate(amp, params, predictions).reshape(1, -1)
            )
        sims = torch.cat(sims, dim=0)

        # Compute composite error
        error = mse_score(target_V, sims) + (1 - abs(correlation_score(target_V, sims)))
        err_pool.append(error)

    predictions = pred_pool[np.argmin(err_pool)]

    run_output_folder_name = f"{constants['run_mode']}_{constants['modfiles_mode']}"
    output_folder = os.path.join(
        constants["output"]["output_folder"], run_output_folder_name
    )
    os.mkdir(output_folder)

    with open(os.path.join(output_folder, "constants.json"), "w") as file:
        json.dump(constants, file, indent=2)

    # Save predictions
    pred_df = pd.DataFrame(
        dict(zip(params, predictions.detach().numpy())), index=[0]
    )
    pred_df.to_csv(os.path.join(output_folder, "pred.csv"))

    save_mse_corr(target_V, constants, predictions, output_folder)

    if constants["outputs"]["produce_plots"]:
        i = 0
        while i < len(constants["optimization_parameters"]["amps"]):
            save_prediction_plots(
                target_V[i].reshape(1, -1),
                constants["amps"][i],
                constants,
                predictions,
                output_folder,
            )
            i += 5

    if os.path.exists("x86_64"):
        os.system("rm -r x86_64")


def run(constants: SimulationConstants):
    p = Process(target=_run, args=[constants])
    p.start()
    p.join()
    p.terminate()


from neuron import h


class CellModel:
    def __init__(self, hoc_file: str, cell_name: str):
        """
        Constructs a Cell object.

        Parameters:
        ----------
        hoc_file: str
            Name of the .hoc file that defines the cell.

        cell_name: str
            Name of the cell in the .hoc file.
        """

        # Load the .hoc file
        h.load_file(hoc_file)

        # Initialize the cell
        hoc_cell = getattr(h, cell_name)()
        self.all = hoc_cell.all
        self.soma = hoc_cell.soma

        # Initialize recorders
        self.t = h.Vector().record(h._ref_t)
        self.Vm = h.Vector().record(self.soma[0](0.5)._ref_v)

        # Init injection
        self.inj = h.IClamp(self.soma[0](0.5))
        self.I = h.Vector().record(h._ref_i)

    def set_parameters(self, parameter_list: list, parameter_values: list) -> None:
        for sec in self.all:
            for index, key in enumerate(parameter_list):
                setattr(sec, key, parameter_values[index])

    def apply_current_injection(self, amp: float, dur: float, delay: float) -> None:
        self.inj.amp = amp
        self.inj.dur = dur
        self.inj.delay = delay
